from .config import AppConfig
import pylunch
import pylunch.utils
from pylunch import lunch

from typing import List

import telegram
from telegram import Update, InlineQueryResultArticle, InputTextMessageContent
from telegram.ext import MessageHandler, CommandHandler, Filters, CallbackContext, \
    InlineQueryHandler, Dispatcher, Updater

from uuid import uuid4
import logging

log = logging.getLogger(__name__)

class CommandHandlers:
    def __init__(self, pylunch_bot: 'PyLunchTelegramBot'):
        self.pylunch_bot = pylunch_bot
        self._commands = {}
    
    def select_instances(self, selectors, fuzzy=False, tags=False, with_disabled=True) -> List[lunch.LunchEntity]:
        
        return self.service.instances.select(selectors, fuzzy=fuzzy, tags=tags, with_disabled=with_disabled)

    def _instance_transform(self, instance):
        content = self.service.resolve_text(instance)
        log.debug(f"[Telegram] entity [{instance.name}]: {content}")
        return f"{instance.display_name} ({instance.name}):\n\n{content}\n\n{instance.url}"

    def print_instances(self, update, instances, transform=None):
        transform = transform if transform is not None else self._instance_transform
        def _write(content): self.send_msg(update, content)
        pylunch.utils.write_instances(instances, transform=transform, writer=_write)
    
    @property
    def service(self) -> lunch.LunchService:
        return self.pylunch_bot.service

    def get_one(self, update: Update, context: CallbackContext):
        result = self.select_instances(context.args)
        log.debug(f"Found for [{context.args}]: {result}")
        self.print_instances(update, result, transform=str)

    def get_menu_one(self, update: Update, context: CallbackContext):
        result = self.select_instances(context.args, tags=False)
        log.debug(f"Found for [{context.args}]: {result}")
        self.print_instances(update, result)

    def get_menu_by_tags(self, update: Update, context: CallbackContext):
        result = self.select_instances(context.args, tags=True)
        log.debug(f"Found for [{context.args}]: {result}")
        self.print_instances(update, result)

    def find_by_tags(self, update: Update, context: CallbackContext):
        result = self.select_instances(context.args, tags=True)
        log.debug(f"Found for [{context.args}]: {result}")
        self.print_instances(update, result, transform=str)

    def list_restaurants(self, update: Update, context: CallbackContext):
        name = " ".join(context.args) if context.args else ""
        result = f"Available ({len(self.service.instances)}):\n"
        for restaurant in self.service.instances.values():
            result += f"{restaurant.name} - {restaurant.url}\n"
        self.send_msg(update, result, markdown=True)

    def get_avaliable_tags(self, update: Update, context: CallbackContext):
        tags = self.service.instances.all_tags()
        result = f"Available ({len(tags)}):\n"
        for tag in tags:
            result += f"- {tag}\n"
        self.send_msg(update, result, markdown=True)

    def start_handler(self, update: Update, context: CallbackContext):
        log.info("[START] Client logged in")
        welcome_message = "Welcome!\n\n"
        welcome_message += self.help_string()
        self.send_msg(update, welcome_message)

    def send_msg(self, update: Update, message, markdown=False):
        log.info(f"[MSG] Message: {message}")
        params = {}
        if markdown:
            params['parse_mode'] = telegram.ParseMode.MARKDOWN
        result = update.message.reply_text(message)
        log.debug(f"[Telegram] Send result: {result}")

    def help_string(self) -> str:
        commands = ""
        for (name, val) in self._commands.items():
            commands += f"- /{name} {val}\n"
        return commands

    def help_handler(self, update: Update, context: CallbackContext):
        log.info("[HELP] Show help")
        result = self.help_string()
        self.send_msg(update, result)

    def inline_query_handler(self, update: Update, context: CallbackContext):
        log.info(f"INLINE initiated")
        query = update.inline_query.query
        log.info(f"[INLINE] Inline query: {query}")
        food_content = InputTextMessageContent(self.service.instances.find_one(query))
        results = [InlineQueryResultArticle(id=uuid4(), title="Get", input_message_content=food_content),]
        update.inline_query.answer(results)

    def error(self, update, context):
        """Log Errors caused by Updates."""
        log.warning('Error "%s" caused by an update %s', context.error, update)
        update.message.reply_text(context.error)

    @property
    def dispatcher(self) -> Dispatcher:
        return self.pylunch_bot.client.dispatcher

    def add_cmd_handler(self, name, help: str, handler, cls=CommandHandler, **kwargs):
        self.add_help(name, help)
        log.info(f"[CMD] Adding command \"{name}\": {help}")
        pass_args = kwargs.get('pass_args', True)
        cmd = cls(name, handler, pass_args=pass_args, **kwargs)
        self.dispatcher.add_handler(cmd)

    def add_help(self, name, help: str):
        self._commands[name] = help


def message_handler(update: Update, context: CallbackContext):
    text = update.message.text
    log.info(f"Received message: {text}")
    update.message.reply_text(f"Pong: {text}")


def register_handlers(bot):
    dispatcher: Dispatcher = bot.client.dispatcher
    cmd = CommandHandlers(bot)
    # Message handler
    dispatcher.add_handler(MessageHandler(Filters.text | Filters.status_update,
                                          message_handler))

    # Basic commands
    cmd.add_cmd_handler('help', 'Show help', cmd.help_handler)
    cmd.add_cmd_handler('start', 'Begin the conversation', cmd.start_handler)

    # Command handler
    cmd.add_cmd_handler('info', '<restaurant> - Get an info for a restaurant', cmd.get_one)
    cmd.add_cmd_handler('tinfo', '<expr> Find restaurants by tags', cmd.find_by_tags)
    cmd.add_cmd_handler('ls', 'Get list of available restaurants', cmd.list_restaurants)
    cmd.add_cmd_handler('tags', 'Get list of all available tags', cmd.get_avaliable_tags)
    cmd.add_cmd_handler('tmenu', '<expr> Get menu for restaurant based on tags', cmd.get_menu_by_tags)
    cmd.add_cmd_handler('menu', '<restaurant> Get menu by restaurant name', cmd.get_menu_one)


    # Inline handler
    dispatcher.add_handler(InlineQueryHandler(cmd.inline_query_handler))
    dispatcher.add_error_handler(cmd.error)


class PyLunchTelegramBot:
    def __init__(self, service: lunch.LunchService):
        self._service = service
        self._client = None

    def run(self):
        token = self.config.telegram_token
        log.info(f"Using the token: {token}")
        register_handlers(self)
        self.client.start_polling()
        self.client.idle()

    @property
    def service(self) -> lunch.LunchService:
        return self._service

    @property
    def config(self) -> AppConfig:
        return self.service.config

    @property
    def client(self) -> Updater:
        if self._client is None:
            self._client = Updater(token=self.config.telegram_token, use_context=True)
        return self._client



import logging
import tempfile
import yaml
from pathlib import Path
from typing import List, Tuple, Mapping
from shutil import copyfile

import click

from pylunch import log_config, lunch, __version__, config, utils

log = logging.getLogger(__name__)

base_dir = Path(__file__).parent.parent
RESOURCES = base_dir / 'resources'
APP_NAME = 'PyLunch'
CONFIG_DIR = click.get_app_dir(APP_NAME.lower())


class CliApplication:
    def __init__(self, config_dir=None):
        self.service: lunch.LunchService = None
        config_dir = config_dir if config_dir is not None else CONFIG_DIR
        self.config_loader = config.YamlLoader(config_dir, 'config.yaml')
        self.restaurants_loader = config.YamlLoader(config_dir, 'restaurants.yaml')

    def init(self, **kwargs) -> 'CliApplication':
        if not self.config_loader.base_dir.exists():
            self._first_run()
        cfg_dict = {**self.config_loader.load(), **kwargs}
        cfg = config.AppConfig(**cfg_dict)
        loaded = self.restaurants_loader.load() or dict(restaurants={})
        unwrapped = loaded.get('restaurants') or loaded
        ent = lunch.Entities(**unwrapped)
        
        self.service = lunch.CachedLunchService(cfg, ent) if cfg.use_cache else lunch.LunchService(cfg, ent)
        return self

    def _first_run(self):
        log.info(f"First run detected, crearing config folder: {self.config_loader.base_dir}")
        self.config_loader.base_dir.mkdir(parents=True)
        self.config_loader.save(data=dict(restaurants='./restaurants.yaml'))
        self.restaurants_loader.save(data={})
        
    def save_restaurants(self):
        log.info("Saving restaurants")
        self.restaurants_loader.save(self.service.instances.to_dict())

    def select_instances(self, selectors, fuzzy=False, tags=False, with_disabled=True) -> List[lunch.LunchEntity]:
        return self.service.instances.select(selectors, fuzzy=fuzzy, tags=tags, with_disabled=with_disabled)
 

pass_app = click.make_pass_decorator(CliApplication)

@click.group(help=f'{APP_NAME} CLI tool')
@click.version_option(version=__version__)
@click.option('-L', '--log-level', help=f'Set log level (d|i|w|e) - default=w', default=None)
@click.option('-C', '--no-cache', help=f'Disable cache', is_flag=True, default=False)
@click.option('-c', '--config-dir', help=f'Location to the configuration directory', default=None)
@click.option('-F', '--format', help='Set output format', default=None)
@click.pass_context
def main_cli(ctx=None, log_level=None, format=None, no_cache=False, config_dir=None, **kwargs):
    log_config.load(log_level)
    app = CliApplication(config_dir=config_dir)
    ctx.obj = app.init(no_cache=no_cache, format=format, log_level=log_level)


@main_cli.command(name='ls', help='List all available restaurants')
@click.option('-l', '--limit', help="Limits number of restaurants to be shown", required=False, default=None)
@pass_app
def cli_list(app: CliApplication, limit=None):
    print(app.service.to_string())
    

@main_cli.command(name='menu', help='Get menu for a restaurant')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@click.option("-U", "--update-cache", help="Update cache entry", default=False, is_flag=True)
@pass_app
def cli_menu(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False, update_cache=False):
    instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags, with_disabled=False)
    if update_cache:
        for item in app.service.clear_cache(instances):
            print(f"Udating cache for: {item}")
    print_instances(app.service, instances)


@main_cli.command(name='info', help='Get info for the restaurant')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_info(app: CliApplication, selectors=None, fuzzy=False, tags=False):
    instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
    print_instances(app.service, instances, transform=str)


@main_cli.command(name='import', help='Import restaurants')
@click.argument('restaurants', nargs=-1)
@click.option("-O", "--override", help="Overide the restaurant if exists", default=False, is_flag=True)
@pass_app
def cli_import(app: CliApplication, restaurants=None, override=False):
    if not restaurants:
        print("Not provided any files to import from")
    for rest_file in restaurants:
        print(f"Importing restaurant: {rest_file}")
        app.service.import_file(rest_file, override=override)
    app.save_restaurants()

@main_cli.command(name='export', help='Export restaurants')
@click.option("-f", "--file", help="Export to file", default=None)
@pass_app
def cli_export(app: CliApplication, file=None):
    if file is None:
        if app.restaurants_loader.full_path.exists():
            content = app.restaurants_loader.full_path.read_text('utf-8')
            print(content)
        else:
            print(f"Error: Restaurants file not exists: {app.restaurants_loader.full_path}")
        
    else:
        # not a hack :-)
        copyfile(src=str(app.restaurants_loader.full_path), dst=file)


@main_cli.command(name='add', help='Adds a new restaurant')
@click.option("-n", "--name", help="Restaurant name", default=False)
@click.option("-d", "--display-name", help="Restaurant name", default=False)
@click.option("-u", "--url", help="Restaurant url", default=False)
@click.option("-s", "--selector", help="Restaurant css selector", default=False)
@click.option("-t", "--tags", help="Restaurant tags", default=False, multiple=True)
@click.option("-p", "--param", help="Additional param", default=False, multiple=True)
@click.option("-O", "--override", help="Overide the restaurant if exists", default=False, is_flag=True)
@pass_app
def cli_add(app: CliApplication, name, display_name, url, selector, tags, params, override=False):
    tags = list(tags) if tags else []        
    config = dict(name=name, url=url, display_name=display_name, tags=tags, selector=selector)
    if params:
        config.update(**_params_dict(params), override=override)
    app.service.instances.register(**config)
    app.save_restaurants()


@main_cli.command(name='rm', help='Removes the restaurant')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_rm(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False):
    instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
    for instance in instances:
        print(f"Removing: {instance.name}")
        del app.service.instances[instance.name]
    app.save_restaurants()


@main_cli.command(name='enable', help='Enables the restaurants')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_enable(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False):
    instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
    for instance in instances:
        if 'disabled' in instance.keys():
            print(f"Enabling instance {instance.name}: {instance}")
            del app.service.instances[instance.name]['disabled']
    app.save_restaurants()


@main_cli.command(name='disable', help='Disables the restaurants')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_disable(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False):
    instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
    for instance in instances:
        print(f"Disabling instance {instance.name}: {instance}")
        app.service.instances[instance.name]['disabled'] = True
    app.save_restaurants()


@main_cli.command(name='edit', help='Edits restaurants DB file')
@pass_app
def cli_edit_restaurants(app: CliApplication):
    cfg = app.restaurants_loader.load()
    content = yaml_edit(cfg)
    if content is None:
        print("No change - not saving")
    else:
        app.restaurants_loader.save(content)


@main_cli.command(name='config', help='Shows the current configuration')
@pass_app
def cli_config(app: CliApplication):
    import yaml
    cfg = app.service.config.config
    print(yaml.safe_dump(dict(Config=cfg)))


@main_cli.command(name='cache-clear', help='Clear a current cache for a day')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_cache_clear(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False):
    if not app.service.config.use_cache:
        print("Not using the cache - no action.")
        return
    
    if isinstance(app.service, lunch.CachedLunchService):
        if not selectors:
            for item in app.service.clear_cache():
                print(f"Clearing: {item}")
        else: 
            instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
            for item in app.service.clear_cache(instances):
                print(f"Clearing: {item}")


@main_cli.command(name='cache-content', help='Show the current cache for a day')
@pass_app
def cli_cache_content(app: CliApplication):
    if not app.service.config.use_cache:
        print("Not using the cache - no action.")
        return

    if isinstance(app.service, lunch.CachedLunchService):
        for (name, cache_paths) in app.service.cache_content().items():
            for path in cache_paths:
                print(f"{name} - {path}")
 
            
@main_cli.command(name='cfg-set', help='Set a config value in the user configuration')
@click.argument('name')
@click.argument('value')
@pass_app
def cli_set_config(app: CliApplication, name, value):
    cfg = app.config_loader.load()
    cfg[name] = value
    app.config_loader.save(cfg)


@main_cli.command(name='cfg-edit', help='Edit a configuration using the editor (Ex: VIM)')
@pass_app
def cli_edit_config(app: CliApplication):
    cfg = app.config_loader.load()
    content = yaml_edit(cfg)
    if content is None:
        print("No change - not saving")
    else:
        app.config_loader.save(content)


@main_cli.command(name='console', help='Start the interactive console (IPython)')
@pass_app
def cli_start_console(app: CliApplication):
    try:
        import IPython
        print("Starting the interactive console using the IPython")
        IPython.embed()
    except ImportError:
        print('\nIPython modeule is not available')

@main_cli.command(name='telegram-bot', help='Start the telegram bot')
@pass_app
def cli_start_console(app: CliApplication):
    try:
        from .telegram_bot import PyLunchTelegramBot
        print("Starting the telegram bot")
        bot = PyLunchTelegramBot(app.service)
        bot.run()
    except ImportError:
        print('\nIPython modeule is not available')


"""
" Helper tools
"""

def yaml_edit(cfg) -> dict:
    content = click.edit(yaml.safe_dump(cfg))
    if content is not None:
        return yaml.safe_load(content)
    return None


def _params_dict(params: List[str]) -> Mapping:
    result = {}
    for param in params:
        (key, val) = param.split('=')
        result[key] = val

def print_instances(service: lunch.LunchService, instances, transform=None):
    transform = transform if transform is not None else lambda x: resolve_menu(service, x)
    utils.write_instances(instances, transform=transform, writer=print)

def resolve_menu(service: lunch.LunchEntity, instance):
    result = _generate_menu_header(instance)
    if service.config.format == 'html':
        result += service.resolve_html(instance)
    else:
        result += service.resolve_text(instance)
    return result 

def _generate_menu_header(instance):
    name_str = f"{instance.display_name} ({instance.name})"
    tags_str = "Tags: " + (", ".join(instance.tags) if instance.tags else '')
    return utils.generate_nice_header(name_str, instance.url, tags_str)


if __name__ == '__main__':
    main_cli()


import flask
import os
import logging
import click
from pathlib import Path
from typing import List, Optional, Mapping

from pylunch import config, lunch, utils

log = logging.getLogger(__name__)


# Find the correct template folder when running from a different location
base_dir = Path(__file__).parent.parent
RESOURCES = base_dir / 'resources'
APP_NAME = 'PyLunch'
CONFIG_DIR = click.get_app_dir(APP_NAME.lower())
tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")

app = flask.Flask(__name__, template_folder=tmpl_dir)


class WebApplication:
    def __init__(self, config_dir=None):
        self.service: lunch.LunchService = None
        config_dir = config_dir if config_dir is not None else CONFIG_DIR
        self.config_loader = config.YamlLoader(config_dir, 'config.yaml')
        self.restaurants_loader = config.YamlLoader(config_dir, 'restaurants.yaml')

    def init(self, **kwargs) -> 'WebApplication':
        if not self.config_loader.base_dir.exists():
            self._first_run()
        cfg_dict = {**self.config_loader.load(), **kwargs}
        cfg = config.AppConfig(**cfg_dict)
        loaded = self.restaurants_loader.load() or dict(restaurants={})
        unwrapped = loaded.get('restaurants') or loaded
        ent = lunch.Entities(**unwrapped)
        
        self.service = lunch.CachedLunchService(cfg, ent) if cfg.use_cache else lunch.LunchService(cfg, ent)
        return self

    def _first_run(self):
        log.info(f"First run detected, crearing config folder: {self.config_loader.base_dir}")
        self.config_loader.base_dir.mkdir(parents=True)
        self.config_loader.save(data=dict(restaurants='./restaurants.yaml'))
        self.restaurants_loader.save(data={})
        
    def save_restaurants(self):
        log.info("Saving restaurants")
        self.restaurants_loader.save(self.service.instances.to_dict())

    def select_instances(self, selectors, fuzzy=False, tags=False, with_disabled=True) -> List[lunch.LunchEntity]:
        return self.service.instances.select(selectors, fuzzy=fuzzy, tags=tags, with_disabled=with_disabled)
 
    @classmethod
    def create(cls) -> 'WebApplication':
        app = cls(config_dir=CONFIG_DIR)
        app.init()
        return app


def parse_request():
    rq = flask.request
    args = rq.args 
    result = dict(selectors=rq.args.getlist('r'), tags=rq.args.getlist('t'))
    return result


@app.route("/menu")
def web_menu():
    args = parse_request()
    tags = args['tags']
    selectors = args['selectors']

    app = WebApplication.create()
    instances = None
    if selectors:
        instances = app.select_instances(instances)
    elif tags:
        instances = app.select_instances(tags, tags=True)
    else:
        instances = app.select_instances(selectors=None)
    content = "\n".join(resolve_menu(app.service, inst) for inst in instances)
    return flask.Response(content, mimetype='text/plain')


###
# Helpers
### 

def resolve_menu(service: lunch.LunchEntity, instance):
    result = _generate_menu_header(instance)
    if service.config.format == 'html':
        result += service.resolve_html(instance)
    else:
        result += service.resolve_text(instance)
    return result 

def _generate_menu_header(instance):
    name_str = f"{instance.display_name} ({instance.name})"
    tags_str = "Tags: " + (", ".join(instance.tags) if instance.tags else '')
    return utils.generate_nice_header(name_str, instance.url, tags_str)



import logging
from typing import List, Optional, Mapping, Tuple, Union, Any, MutableMapping, Mapping

import html2text
import requests
import yaml
import json
import datetime
import shutil
import collections
import io
import re
import locale
from bs4 import BeautifulSoup, Tag
from requests import Response
from pyzomato import Pyzomato

from fuzzywuzzy import fuzz, process
from pathlib import Path

from .tags_evaluator import TagsEvaluator
from .config import AppConfig, YamlLoader

from pdfminer import high_level
import pdfminer.layout

log = logging.getLogger(__name__)


class LunchEntity(collections.MutableMapping):
    def __init__(self, config: Mapping[str, Any]):
        self._config = {**config}

    def __getitem__(self, k):
         return self._config.get(k)

    def __setitem__(self, k, v):
         self.config[k] = v
    
    def __delitem__(self, k):
         del self.config[k]
    
    def __iter__(self):
        return iter(self._config)
    
    def __len__(self):
         return len(self.config)

    @property
    def config(self) -> MutableMapping['str', Any]:
        return self._config

    @property
    def resolver(self) -> str:
        return self.config.get('resolver', 'default')

    @property
    def name(self) -> str:
        return self.config.get('name')

    @property
    def url(self) -> str:
        return self.config.get('url')

    @property
    def selector(self) -> str:
        return self.config.get('selector')

    @property
    def request_params(self) -> List:
        return self.config.get('request_params', {})

    @property
    def display_name(self) -> str:
        return self.config.get('display_name') or self.name

    @property
    def tags(self) -> List[str]:
        return self.config.get('tags')

    @property
    def disabled(self) -> bool:
        return self.config.get('disabled', False)

    @property
    def days(self) -> List[str]:
        return self.config.get('days')

    @property
    def filters(self) -> str:
        return self.config.get('filters')

    def __str__(self) -> str:
        result = f"\"{self.name}\" -"

        if self.display_name:
            result += f" ({self.display_name})"
        if self.tags:
            result += f" {self.tags}"
        
        result += f" {self.url}"

        if self.selector:
            result += f" ({self.selector})"
        if self.request_params:
            result += f" - req_params={self.request_params}"

        if self.resolver and self.resolver != 'default':
            result += f' resolver={self.resolver}'
        return result

    def __repr__(self) -> str:
        return str(self.config)

class AbstractResolver:
    def __init__(self, service: 'LunchService', entity: LunchEntity):
        self.service = service
        self.entity = entity

    def resolve_text(self) -> str:
        return None

    def resolve_html(self) -> str:
        return None

class LunchContentFilter:
    def __init__(self, service: 'LunchService', entity: LunchEntity):
        self.entity = entity
        self.service = service

    def filter(self, content: str) -> str:
        return content

class DayResolveFilter:
    CZ_DAYS = ['Pondelí', 'Úterý', 'Středa', 'Čtvrtek', 'Pátek', 'Sobota', 'Neděle']
    EN_DAYS = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']

    def __init__(self, service: 'LunchService', entity: LunchEntity):
        self.entity = entity
        self.service = service

    @property
    def _week_day(self) -> int:
        return datetime.datetime.today().weekday()

    def _find_pos_content(self, day_num, content, days):
        if day_num >= len(days):
            log.warning(f"Week day \"{day_num}\" is greater than available days {len(days)}")
            return None

        day = days[day_num]
        found = re.search(day, content, re.IGNORECASE)
        if not found:
            log.warning(f"Day \"{day}\" was not found in the content")
            fallback = self._find_pos_content(day_num, content, self.__class__.EN_DAYS)
            if fallback:
                log.info(f"Fallback option has been found: {fallback} position")
                return fallback
            else:
                log.warning(f"Fallback failed for {self.entity.name} and day number {day_num}.")
                return None
        return found.start()
        

    def _parse_by_day(self, content: str, days: List[str]) -> str:
        if not days:
            log.debug("[FILTER] No days selected - sending original content")
            return content
        start = self._find_pos_content(self._week_day, content, days)
        if start is None:
            log.warning("[FILTER] Unable to find begging of the day, sending original conent")
            return content
        end = self._find_pos_content(self._week_day + 1, content, days)
        if end is None:
            log.debug(f"[FILTER] Sending content from {start} to the end of the document.")
            return content[start:]
        log.debug(f"[FILTER] Sending content from {start} to {end}.")
        return content[start:end]


    def _select_correct_days(self, content: str):
        days = self.entity.days
        if days:
            return days
        
        today_num = self._week_day
        options = [self.__class__.CZ_DAYS, self.__class__.EN_DAYS]
        for option in options:
            if re.search(option[today_num], content, re.IGNORECASE):
                log.debug(f"[DAYS] Selecting correct days for [{option[today_num]}]: {option}")
                return option
        return None

    def filter(self, content: str):
        days = self._select_correct_days(content)
        result = self._parse_by_day(content, days)
        return result


class LunchResolver(AbstractResolver):
    @property
    def request_url(self) -> str:
        return self.entity.url

    def _make_request(self, **kwargs) -> requests.Response:
        response = requests.get(self.request_url, **kwargs)
        if not response.ok:
            log.warning(f"[LUNCH] Error[{response.status_code}] ({self.entity.name}): {response.content}")
            print(f"Error[{response.status_code}] {self.entity.name}: ", response.content)
        else:
            log.debug(f"[RES] Response [{response.status_code}] {self.entity.name}: {response.content}")
        return response

    def _parse_response(self, response: Response) -> List[Tag]:
        soap = BeautifulSoup(response.content, "lxml")
        sub = soap.select(self.entity.selector) if self.entity.selector else soap
        log.info(f"[LUNCH] Parsed[{self.entity.name}]: {sub}")
        return sub

    def resolve_text(self) -> str:
        html_string = self.resolve_html()
        if html_string is None:
            return None
        return to_text(html_string)

    def resolve_html(self) -> str:
        response = self._make_request(**(self.entity.request_params or {}))
        parsed = self._parse_response(response=response)
        content = self.to_string(parsed)
        if not content:
            log.warning(f"[HTML] Content is empty for {self.entity.name} - {self.entity.url} ({self.entity.selector})")
            return None
        else:
            log.debug(f"[HTML] Extracted content {self.entity.name}: {content}")
        return content 
        
    @classmethod
    def to_string(cls, parsed) -> str:
        if isinstance(parsed, list):
            items = [str(item) for item in parsed]
            return "".join(items)
        else:
            return str(parsed.extract())


class ZomatoResolver(AbstractResolver):
    ZOMATO_NOT_ENABLED="""Zomato key is not set, please get the zomato key 
                        and set add it to the configuration as property `zomato_key`."""
    @property
    def zomato(self) -> Pyzomato:
        return self.service.zomato

    def resolve_json(self):
        if self.zomato is None:
            return None
        content = self.zomato.getDailyMenu(self.entity.selector)
        log.info(f"[ZOMATO] Response: {json.dumps(content, indent=2)}")
        return content
    
    def resolve_html(self) -> str:
        resolved = self.resolve_json()
        if resolved is None:
            return f"<p>{ZomatoResolver.ZOMATO_NOT_ENABLED}</p>"
        result = "<div>\n"
        lines = self._make_lines(resolved)
        for line in lines:
            result += f"<p>{line}</p>\n"
        return result + "\n</div>"

    def resolve_text(self) -> str:
        content = self.resolve_json()
        if content is None:
            return ZomatoResolver.ZOMATO_NOT_ENABLED
        return "\n".join(self._make_lines(content))

    def _make_lines(self, content: dict) -> list:
        result = []
        menus = content['daily_menus']
        for menu in menus:
            menu = menu.get('daily_menu')
            dishes = menu['dishes']
            for dish in dishes:
                dish = dish['dish']
                result.append(f"{dish['name']} - {dish['price']}")
        return result


class PDFResolver(AbstractResolver):
    def resolve_pdf(self):
        response = requests.get(self.entity.url)
        if not response.ok:
            log.error(f"Unnable to get response from: {self.url}")
            return None
        text = self._resolve_text_from_content(io.BytesIO(response.content))
        log.info(f"[PDF] Resolved text: {text}")
        return text
    
    def resolve_html(self) -> str:
        result = "<div>\n"
        result += f'<a href="{self.entity.url}">{self.entity.display_name}<a>'
        return result + "\n</div>"

    def resolve_text(self) -> str:
        text = self.resolve_pdf()
        return f"PDF is available at: {self.entity.url}\n\n{text}" 

    def _resolve_text_from_content(self, stream: io.BytesIO):
        out = io.BytesIO()
        laparams = pdfminer.layout.LAParams()
        for param in ("all_texts", "detect_vertical", "word_margin", "char_margin", "line_margin", "boxes_flow"):
            paramv = locals().get(param, None)
            if paramv is not None:
                setattr(laparams, param, paramv)

        high_level.extract_text_to_fp(stream, outfp=out, laparams=laparams)
        return out.getvalue().decode('utf-8')

    
class LunchCollection(collections.MutableMapping):
    def __init__(self, cls_wrap=None, **kwargs):
        self._collection = { key: cls_wrap(val) if cls_wrap else val for (key, val) in kwargs.items() } 

    @property
    def collection(self) -> MutableMapping[str, Any]:
        return self._collection

    def __getitem__(self, k):
         return self.collection.get(k)
    def __setitem__(self, k, v):
         self.collection[k] = v
    def __delitem__(self, k):
         del self.collection[k]
    def __iter__(self):
        return iter(self.collection)
    def __len__(self):
         return len(self.collection) 

class Resolvers(LunchCollection):
    def register(self, name: str, cls: type):
        log.info(f"[ADD] Resolver [{name}]: {cls.__name__}")
        self._collection[name] = cls

    def get(self, name: str) -> type:
        return self._collection.get(name, LunchResolver)

    def for_entity(self, entity: LunchEntity) -> LunchResolver:
        return self.get(entity.resolver)

class Filters(LunchCollection):
    def register(self, name: str, cls: type):
        log.info(f"[ADD] Filter [{name}]: {cls.__name__}")
        self._collection[name] = cls

    def get(self, name: str) -> type:
        return self._collection.get(name, LunchContentFilter)

    def for_entity(self, entity: LunchEntity) -> LunchContentFilter:
        log.debug(f"[FILTER] Filters for entity {entity.name} ~> {entity.filters}")
        return [ self.get(flt) for flt in (entity.filters or []) ]


class Entities(LunchCollection):
    def __init__(self, **kwargs):
        super().__init__(cls_wrap=LunchEntity, **kwargs)
    @property
    def entities(self) -> MutableMapping[str, LunchEntity]:
        return self.collection

    def __getitem__(self, name) -> Optional[LunchEntity]:
        if name in self.entities.keys():
            instance = self.entities[name]
            log.info(f"[LUNCH] Found in entities {name}: {instance}")
            return instance
        else:
            log.warning(f"[LUNCH] Not found in entities {name}")
            return None

    def __setitem__(self, name, config):
        self.register(name=name, **config)

    def find_one(self, name: str):
        return self.get(name) or self.fuz_find_one(name)[0]

    def find_all(self, name: str, limit=10):
        return [i[0] for i in self.fuz_find(name, limit)]

    def fuz_find(self, name: str, limit=10) -> List[Tuple]:
        return process.extract(name, self.entities,
                               processor=lambda x: x if isinstance(x, str) else x.name,
                               scorer=fuzz.token_sort_ratio,
                               limit=limit)

    def fuz_find_one(self, name: str) -> Tuple:
        return process.extractOne(name, self.entities,
                                  processor=lambda x: x if isinstance(x, str) else x.name,
                                  scorer=fuzz.token_sort_ratio)

    def register(self, name: str, url: str, display_name: str=None, tags=None, 
                    selector=None, request_params=None, override=False, **kwargs):
        if name in self.entities.keys():
            if override:
                log.info(f"[REG] Overriding already existing: {name}")
            else:
                log.info(f"[REG] Skipping {name} since it already exists.")
                return
        config = dict(name=name, url=url, selector=selector, display_name=display_name, tags=tags, request_params=request_params, **kwargs)
        instance = LunchEntity(config)
        log.info(f"[REG] Register [{name}]: {instance}")
        self.entities[name] = instance

    def all_tags(self) -> List[str]:
        accumulator = set()
        for entity in self.entities.values():
            accumulator.update(entity.tags if entity.tags else {})
        return list(accumulator)
    
    def find_by_tags(self, expression: str):
        tags = TagsEvaluator(expression, self.all_tags())
        result = [ entity for entity in self.entities.values() if tags.evaluate(entity.tags) ]
        log.info(f"[FIND] Found by tags {expression}: {result}")
        return result

    def to_dict(self) -> dict:
        return { 'restaurants': { name: value.config for (name, value) in self.collection.items() } }

    def select(self, selectors, fuzzy=False, tags=False, with_disabled=True) -> List[LunchEntity]:
        def _get() -> List['lunch.LunchEntity']:
            if selectors is None or len(selectors) == 0:
                return list(self.values())
            if tags:
                full = " ".join(selectors)
                return self.find_by_tags(full)
            if fuzzy:
                return [ self.fuz_find_one(select) for select in selectors ]
            return [ self.find_one(select) for select in selectors ]

        instances = _get()
        instances = [instance for instance in instances if instance is not None]
        if with_disabled:
            return instances

        return [ item for item in instances if item and not item.disabled]
        

class LunchService:
    def __init__(self, config: AppConfig, entities: Entities):
        self._entities: Entities = entities
        self._resolvers: Resolvers = Resolvers(default=LunchResolver, zomato=ZomatoResolver, pdf=PDFResolver)
        self._filters: Filters = Filters(raw=LunchContentFilter, day=DayResolveFilter)
        self._config: AppConfig = config
        self._zomato: Pyzomato = None

    @property
    def zomato(self) -> Pyzomato:
        if self._zomato is None:
            if self.config.zomato_key is None:
                return None
            self._zomato = Pyzomato(self.config.zomato_key)
        return self._zomato
    
    @property
    def config(self) -> AppConfig:
        return self._config

    @property
    def resolvers(self) -> Resolvers:
        return self._resolvers

    @property
    def filters(self) -> Filters:
        return self._filters

    @property
    def instances(self) -> Entities:
        return self._entities
  
    def import_file(self, file: Tuple[Path, str], override=False):
        file = Path(file)
        if not file.exists():
            log.warning(f"[IMPORT] File not exists: {file}")
            return 
        with file.open("r", encoding='utf-8') as fp:
            log.info(f"[IMPORT] Importing file: {file}")
            restaurants = yaml.safe_load(fp)
            for (name, restaurant) in restaurants['restaurants'].items():
                restaurant['name'] = name
                restaurant['override'] = override
                if restaurant.get('tags') and restaurant.get('resolver', 'default') != 'default' and restaurant['resolver'] not in restaurant['tags']:
                    restaurant['tags'].append(restaurant['resolver'])
                self.instances.register(**restaurant)  
                    
    def process_lunch_name(self, name: str) -> str:
        if not name or name == 'list':
            return self.to_string()
        log.info(f"[CMD] Lunch name: {name}")
        instance = self.instances.get(name)

        try:
            content = f"Restaurant: \"{name}\" - {instance.url}\n" + instance.invoke()
            log.debug(f"Content: {content}")
            return content
        except Exception as ex:
            return "ERR: {ex}"

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        result = f"Available ({len(self.instances)}): \n"
        for restaurant in self.instances.values():
            result += f" - {restaurant.name} - {restaurant.url}\n"
        return result

    def resolve_text(self, entity: LunchEntity) -> str:
        resolver = self.resolvers.for_entity(entity)
        log.debug(f"[RESOLVER] Using the resolver: {resolver.__name__}")
        content = resolver(self, entity).resolve_text()
        if not content:
            log.warning(f"[SERVICE] No content for {entity.name}")
            return None
        filters = self.filters.for_entity(entity)
        for flt in filters:
            log.debug(f"[FILTER] Using the text filter: {flt.__name__}")
            content = flt(self, entity).filter(content)
        return content
    
    def resolve_html(self, entity: LunchEntity) -> str:
        resolver = self.resolvers.for_entity(entity)
        return resolver(self, entity).resolve_html()


class CachedLunchService(LunchService):
    def __init__(self, cfg: AppConfig, entities: Entities):
        super().__init__(cfg, entities)
        self.cache_base = Path(self.config.cache_dir)

    def resolve_text(self, entity: LunchEntity) -> str:
        return self._resolve_any(entity, super().resolve_text, ext='txt')

    def resolve_html(self, entity: LunchEntity) -> str:
        return self._resolve_any(entity, super().resolve_html, ext='html')

    def _resolve_any(self, entity: LunchEntity, func, ext: str = None):
        file = self._entity_file(entity, ext=ext)
        if file is not None and file.exists():
            log.debug(f"[CACHE] Cache hit for {entity.name}: {file}")
            return file.read_text(encoding='utf-8')

        content = func(entity)
        if not content:
            log.warning(f"[CACHE] No content provided for {entity.name}")
            return f'No content provided for {entity} - not caching'
        if self.cache_base is not None:
            self._create_cache_for_day()
            log.info(f"[CACHE] Writing \"{entity.name}\" to cache: {file}")
            file.write_text(str(content), encoding='utf-8')
        return content

    def _create_cache_for_day(self, day: str=None) -> Path:
        """
        Expected format: YYYY-MM-DD
        """
        if not self._cache().exists():
            self._cache().mkdir(parents=True)
        return self._cache

    @property
    def _today_date(self) -> str:
        return datetime.datetime.today().strftime('%Y-%m-%d')

    def _cache(self, date=None) -> Optional[Path]:
        if self.cache_base is None:
            return None
        return self.cache_base / (self._today_date if date is None else date)

    def _entity_file(self, entity: LunchEntity, ext=None) -> Path:
        ext = ext if ext is not None else 'txt'
        if self.cache_base is None:
            return None
        return self._cache() / f"{entity.name}.{ext}"

    def clear_cache(self, lst: List[LunchEntity] = None, full=False):
        cache_dir = self._cache() if not full else self.cache_base
        if not lst:
            log.info(f"Cleaning cache: {cache_dir}")
            shutil.rmtree(str(cache_dir), True)
            return [cache_dir]
        else:
            return self._clear_items(cache_dir, lst)
    
    def _clear_items(self, base_dir, lst) -> List:
        result = []
        for item in lst:
            paths: List[Path] = base_dir.glob(f"{item.name}.*")
            for path in paths:
                log.info(f"[CACHE] Removing {item.name}: {path}")
                path.unlink()
                result.append(path)
        return result

    def cache_content(self, lst: List[LunchEntity] = None) -> Mapping:
        base = self._cache()
        log.debug(f"[CACHE] Base: {base}")
        result =  { name: list(str(pth) for pth in base.glob(f'{entity.name}.*')) for (name, entity) in self.instances.items() }
        log.debug(f"[CACHE] Content: {result}")
        return result
        


def to_text(content):
    h = html2text.HTML2Text()
    h.ignore_links = True
    h.ignore_images = True
    h.ignore_tables = True
    h.ignore_emphasis = True
    return h.handle(str(content)).strip()



import logging

log = logging.getLogger(__name__)


def write_instances(instances, transform=None, writer=None):
    writer = writer or print
    log.debug(f"Printing: {instances}")
    if instances is None or not instances:
        writer("**No instance has been found**")

    elif isinstance(instances, list):
        for instance in instances:
            if instance is not None:
                log.info(f"Sending one response from list: {instances}")
                writer(transform(instance))
    else:
        log.info(f"Sending one response: {instances}")
        writer(transform(instances))

def generate_nice_header(*strings):
    def _for_print(max_l, curr, char='='):
        return char * (max_l - curr)

    def _print_text(max_l, text):
        buffer = ''
        buffer += f"\n===  {text}"
        buffer +=_for_print(max_l, len(text), char=' ') 
        buffer += "  ==="
        return buffer

    def _beg_end_line(max_l):
        return f"\n{_for_print(max_l + 10, 0)}"
    
    max_len = max(len(text) for text in strings)

    result = _beg_end_line(max_len)
    for text in strings:
        result += _print_text(max_len, text)
    result += _beg_end_line(max_len)
    return result + "\n\n"



import io
import re
from pathlib import Path

from setuptools import find_packages, setup

with io.open('pylunch/__init__.py', 'rt', encoding='utf8') as f:
    VERSION = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

long_description = Path('README.md').read_text(encoding='utf-8')

requirements = ['pyyaml', 
                'click', 
                'requests', 
                'coloredlogs', 
                'python-Levenshtein', 
                'fuzzywuzzy', 
                'html2text', 
                'beautifulsoup4', 
                'lxml',
                'pyzomato',
                'pdfminer-six',
                'ipython',
                'flask',
                ]
                
entry_points = {'console_scripts': ['pylunch = pylunch.cli:main_cli', ]}
extra_requirements = {
    'dev': ['pytest>=3', 'coverage' ],
    'docs': ['sphinx', ]
}

setup(name='pylunch',
      version=VERSION,
      description='Internal pylunch cli tool to get lunch info',
      author='Peter Stanko',
      author_email='peter.stanko0@gmail.com',
      url='https://gitlab.com/pestanko/pylunch',
      packages=find_packages(exclude=("tests",)),
      long_description=long_description,
      long_description_content_type='text/markdown',
      include_package_data=True,
      install_requires=requirements,
      extras_require=extra_requirements,
      entry_points=entry_points,
      classifiers=[
          "Programming Language :: Python :: 3",
          'Programming Language :: Python :: 3.7',
          "Operating System :: OS Independent",
          "License :: OSI Approved :: Apache Software License",
          'Intended Audience :: Developers',
          'Topic :: Utilities',
      ], )
